/*  pusher.c

    Copyright (c) 1993 by Eugene Crosser
    Uses curses module from ELM 2.4 (slightly hacked)
    
    Permission is granted for free distribution and use of this text
    provided the copyright notice is preserved.

    This is an implementation of "Sokoban" video game (originally
    from Yamacha computers). 

    Development of this program was inspired by Andy Nikolaev, author
    of an MS/DOS version - thanx Andy!

    Eugene Crosser <crosser@pccross.msk.su>, 2:5020/23.14@FidoNet
*/

#include "curses.h"
#include <stdio.h>
#include <malloc.h>
#include <signal.h>
#include <ctype.h>
#include <string.h>

#define EMPTY 0x7f
#define SPACE 0
#define WALL 1
#define BOX 2
#define ME 3
#define MASK 0x80

int COLS,LINES;

#define MAXMAZE 60
int mazeno=1;
char *board=NULL;
char board_x,board_y;
int me_x,me_y;
int boxes,in_place;
int running=1;

int u_me_x,u_me_y,u_box_x,u_box_y,u_space_x,u_space_y,u_inpl,can_undo=0;

main()
{
	FILE *mazefp;
	char mazenm[128];

	init_video();

	while (running)
	{
		sprintf(mazenm,"/usr/local/lib/games/pusher/maze%02d.dat",
			mazeno);
		if ((mazefp=fopen(mazenm,"r")) == NULL)
			error("cannot read maze file");
		fread(&board_x,1,1,mazefp);
		fread(&board_y,1,1,mazefp);
		board=(char *)malloc(board_x*board_y);
		if (!board)
			error("Not enough RAM for playboard");
		fread(board,board_x,board_y,mazefp);

		play();

		free(board);
		fclose(mazefp);
	}

	close_video();

	return(0);
}

error(str)
char *str;
{
	close_video();
	fprintf(stderr,"%s\n",str);
	exit(1);
}

play()
{
	int c;

	int quit=0;
	redraw();
	while (!quit)
	{ 
		switch (c=tgetch())
		{
		case 'I':
		case 'i':       info(); break;

		case 'U':
		case 'u':	undo(); break;

		case 'N':
		case 'n':	quit=1; if (mazeno < 60) mazeno++;
				can_undo=OFF; break;

		case 'P':
		case 'p':	quit=1; if (mazeno > 1) mazeno--;
				can_undo=OFF; break;

		case 'S':
		case 's':	quit=1; can_undo=OFF; break;

		case 'Q':
		case 'q':	quit=1; running=0; break;

		case 12:	redraw(); break; /* Ctrl-L */

		case KEY_UP:
		case 'K':
		case 'k':	moveme(-1,0); break;

		case KEY_DOWN:
		case 'J':
		case 'j':	moveme(1,0); break;

		case KEY_LEFT:
		case 'H':
		case 'h':	moveme(0,-1); break;

		case KEY_RIGHT:
		case 'L':
		case 'l':	moveme(0,1); break;

		default:	/*printf("got %d, no match",c);*/ break;
		}
	MoveCursor(LINES,0);
	}	
}

redraw()
{
	int i,j;
	char buf[128];
	char *title="Pusher level %d";
	char *footer1="H,J,K,L, Arrow keys - move; Ctrl-L - redraw; U)ndo";
	char *footer2="I)nfo; N)ext / P)revious / S)ame level; Q)uit";

	ScreenSize(&LINES,&COLS);
	ClearScreen();

	sprintf(buf,title,mazeno);
	cline(0,buf);
	cline(LINES-1,footer1);
	cline(LINES,footer2);

	boxes=0;
	in_place=0;

	for (j=0;j<board_y;j++) for (i=0;i<board_x;i++)
	{
		if ((board[i+j*board_x] & 0x7f) == ME)
		{
			me_x=i;
			me_y=j;
		}
		if ((board[i+j*board_x] & 0x7f) == BOX)
		{
			boxes++;
			if (board[i+j*board_x] & 0x80) in_place++;
		}
		show(j,i);
	}
	MoveCursor(LINES,0);
}

info()
{
  ClearScreen();

  cline(2,"Pusher for UNIX");
  cline(4,"Copyright (c) 1993 by Eugene Crosser");
  cline(6,"Maze configurations from original Sokoban");
  cline(7,"Curses module from ELM 2.4");
  cline(10,"What you need:");
  cline(12,"Move all boxes ( [] ) to hatched cells ( ,' )");
  cline(13,"You control the guy ( <> ) who can push");
  cline(14,"only one box at a time.");

  cline(LINES,"Press any key to continue...");

  tgetch();
  redraw();
}

cline(ln,str)
int ln;
char *str;
{
  int i,l;

  l=strlen(str);
  if ((l <= COLS) && (ln <= LINES))
    {
      i=(COLS-l)/2;
      PutLine(ln,i,str);
    }
}

show(j,i)
int j,i;
{
	int x,y;
	int mask;

	x=i*2+(2*COLS/2-(board_x*2))/2;
	y=j+(LINES-board_y)/2;

	mask=board[i+j*board_x] & MASK;
	if ((x >= 0) && (x < COLS) && (y >= 0) && (y < LINES))
	switch(board[i+j*board_x] & 0x7f)
	{
		case SPACE:	if (mask) PutLine(y,x,",'");
				else PutLine(y,x,"  "); break;

		case EMPTY:
		case WALL:	StartInverse();
				PutLine(y,x,"::");
				EndInverse();
				break;

		case BOX:	if (mask)
				{
					StartInverse();
					PutLine(y,x,"{}");
					EndInverse();
				}
				else PutLine(y,x,"[]");
				break;

		case ME:	if (mask) StartInverse();
				PutLine(y,x,"<>");
				if (mask) EndInverse();
				break;

		default:	PutLine(y,x,"??"); break;
	}
}

moveme(dy,dx)
int dy,dx;
{
	int newx,newy;

	newx=me_x+dx;
	newy=me_y+dy;
	u_inpl=in_place;
	if (newx < 0) return(0);
	if (newy < 0) return(0);
	if (newx > board_x) return(0);
	if (newy > board_y) return(0);
	if ((board[newx+newy*board_x] & 0x7f) == WALL) return(0);
	if ((board[newx+newy*board_x] & 0x7f) == BOX)
	{
		if (!movebox(newy+dy,newx+dx)) return(0);
		u_box_x=newx;
		u_box_y=newy;
		u_space_x=newx+dx;
		u_space_y=newy+dy;
		if ((board[newx+newy*board_x] & 0x80) &&
		    !(board[newx+dx+(newy+dy)*board_x] & 0x80))
			in_place--;
		if (!(board[newx+newy*board_x] & 0x80) &&
		    (board[newx+dx+(newy+dy)*board_x] & 0x80))
			in_place++;
	}
	else
	{
		u_box_x=-1;
		u_box_y=-1;
		u_space_x=newx;
		u_space_y=newy;
	}
	board[me_x+me_y*board_x] &= 0x80;
	show(me_y,me_x);
	board[newx+newy*board_x] &= 0x80;
	board[newx+newy*board_x] |= ME;
	u_me_x=me_x;
	u_me_y=me_y;
	can_undo=ON;
	me_x=newx;
	me_y=newy;
	if (boxes == in_place)
	{
		board[me_x+me_y*board_x] &= 0x80;
		leveldone();
	}
	show(me_y,me_x);
}

movebox(newy,newx)
int newy,newx;
{
	if (newx < 0) return(0);
	if (newy < 0) return(0);
	if (newx > board_x) return(0);
	if (newy > board_y) return(0);
	if ((board[newx+newy*board_x] & 0x7f) == SPACE)
	{
		board[newx+newy*board_x] &= 0x80;
		board[newx+newy*board_x] |= BOX;
		show(newy,newx);
		return(1);
	}
	return(0);
}

leveldone()
{
	int i,j;
	char *greet=" * * *  L E V E L   D O N E  * * * ";

	cline(2,greet);
}

undo()
{
	if (!can_undo) return(0);
	can_undo=OFF;
	me_x=u_me_x;
	me_y=u_me_y;
	in_place=u_inpl;
	board[u_me_x+u_me_y*board_x] &= 0x80;
	board[u_me_x+u_me_y*board_x] |= ME;
	if (in_place != boxes) show(u_me_y,u_me_x);
	board[u_space_x+u_space_y*board_x] &= 0x80;
	board[u_space_x+u_space_y*board_x] |= SPACE;
	show(u_space_y,u_space_x);
	if (u_box_x != -1)
	{
		board[u_box_x+u_box_y*board_x] &= 0x80;
		board[u_box_x+u_box_y*board_x] |= BOX;
		show(u_box_y,u_box_x);
	}
}

#ifdef SIGWINCH
void sigwinch(sig)
int sig;
{
	redraw();
	signal(SIGWINCH,sigwinch);
}
#endif

init_video()
{
	InitScreen();
	Raw(ON);
	UseKeypad(ON);
#ifdef SIGWINCH
	signal(SIGWINCH,sigwinch);
#endif
}

close_video()
{
	MoveCursor(LINES,COLS);
	Raw(OFF);
#ifdef SIGWINCH
	signal(SIGWINCH,SIG_DFL);
#endif
	printf("\n");
}

tgetch()
{
	int c;

	c=ReadCh();
	/* printf("\rread: '%c' (%d)\n",c,c);fflush(stdout); */
	return(c);
}
